from skyciv.classes.model.components.nodes.node import Node
from skyciv.utils.helpers import next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent


class Nodes(ModelCollectionComponent):
    """Creates an instance of the SkyCiv Nodes class.
    """

    def add(self, x: int, y: int, z: int) -> int:
        """Create a node with the next available ID.

        Args:
            x (int): The X coordinate of the node.
            y (int): The Y coordinate of the node.
            z (int): The Z coordinate of the node.

        Returns:
            int: The ID of the created node.
        """
        next_index = next_object_key(self)
        nodeId = self.id_from_coords(x, y, z)

        if (nodeId != None):
            print('Prevented overwriting an existing node with the Nodes.add() method.')
            return nodeId
        else:
            node = Node(x, y, z)
            setattr(self, str(next_index), node)
            return next_index

    def id_from_coords(self, x: int, y: int, z: int) -> int:
        """Find a node's ID from its coordinates.

        Args:
            x (int): The x coordinate of the node whose ID to find.
            y (int): The y coordinate of the node whose ID to find.
            z (int): The z coordinate of the node whose ID to find.

        Returns:
            int: The ID of the found node.
        """
        id = None

        for k, v in vars(self).items():
            if (v.x == x and v.y == y and v.z == z):
                id = k

        return id



from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent
from skyciv.classes.model.components.area_loads.area_load import AreaLoad
from skyciv.utils.helpers import keyvals, next_object_key
from typing import List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

class AreaLoads(ModelCollectionComponent):
    """Creates an instance of the SkyCiv AreaLoads class.
    """

    def add(self,
            type: Literal["one_way", "two_way", "column_wind_load", "open_structure"],
            nodes: List[int],
            mag: float,
            direction: Literal["X", "Y", "Z"],
            elevations: str = 0,
            mags: str = 0,
            column_direction: str = None,
            loaded_members_axis: Literal["all", "major"] = None,
            LG: str = None
            ) -> int:
        """Create an area load with the next available ID.

        Args:
            type (str): How the area load should distribute the load. {"one_way" | "two_way" | "column_wind_load" | "open_structure"}.
            nodes (list[int]): The IDs of the nodes which define the area for loading. Specify 3 or 4 values, in sequential order(clockwise or counterclockwise direction).
            mag (float): The magnitude of the load, in the units of pressure.
            direction (str): The direction of the load in the global axes. {"X" | "Y" | "Z"}.
            elevations (str, optional): Relevant only if "type": "column_wind_load". The elevations between which the corresponding pressure magnitudes (see next row in this table) should be applied. This property should have 1 more value than the corresponding pressure magnitudes property. Defaults to 0.
            mags (str, optional): Relevant only if "type": "column_wind_load". The magnitudes of pressures which should be applied between the corresponding elevations(see above row in this table). This property should have 1 less value than the corresponding elevations property. Defaults to 0.
            column_direction (str, optional): Relevant only if "type": "one_way" or "type": "column_wind_load". The span direction of the applied area load. The values must be the IDs of 2 nodes which are in the nodes property. Defaults to None.
            loaded_members_axis (str, optional): Relevant only if "type": "open_structure".Whether to apply the open structure load to all members attaching to the nodes(indicated by all), or to only those members which lie along the global XYZ axes(indicated by major). Defaults to None. {"all" | "major"}.
            LG (str, optional): The load group to which this area load belongs. Defaults to None.

        Returns:
            int: The ID of the new load.
        """

        next_index = next_object_key(self)
        element_ids = self.get_area_load_ids_from_node_ids(nodes)

        if (element_ids != None):
            print('There is more than one area load with the same nodes.')

        al = AreaLoad(
            type,
            nodes,
            mag,
            direction,
            elevations,
            mags,
            column_direction,
            loaded_members_axis,
            LG
        )
        setattr(self, str(next_index), al)

        return next_index

    def get_area_load_ids_from_node_ids(self, nodes: List[int]) -> List[int]:
        """Get the IDs of all area loads that match the provided nodes array. Node order IS considered.

        Args:
            nodes (list[int]): An array of node IDs that define the area load. At least 3 IDs are required.

        Returns:
            list[int]: An array of member IDs or None if none exist.
        """
        ids = []

        for k, v in keyvals(self):
            if v.nodes == nodes:
                ids.append(k)

        if len(ids) == 0:
            ids = None

        return ids



from skyciv.classes.model.components.self_weights.self_weight import SelfWeight
from skyciv.utils.helpers import next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent


class SelfWeights(ModelCollectionComponent):
    """Creates an instance of the SkyCiv SelfWeights class.
    """

    def add(self, enabled: bool = True, x: int = 0, y: int = 0, z: int = 0, LG: str = 'SW1') -> int:
        """Add an instance of self-weight.

        Args:
            enabled (bool, optional): Defaults to True.
            x (int, optional): Acceleration due to gravity in the x-axis, defined as a multiplier of the gravitational constant g. Defaults to 0.
            y (int, optional): Acceleration due to gravity in the y-axis, defined as a multiplier of the gravitational constant g. Defaults to 0.
            z (int, optional): Acceleration due to gravity in the z-axis, defined as a multiplier of the gravitational constant g. Defaults to 0.
            LG (str, optional): The load group that the self-weight belongs to. Defaults to existing value. Defaults to 'SW1'.

        Example:
            For a simple model where +Y on the global axis is vertical,
            the following will apply [1 ✕ gravity] for the model self-weight::

            sws = SelfWeights()
            sws.add(y=-1, LG="SW1")

        Returns:
            int: The ID of the newly created self-weight.
        """
        next_index = next_object_key(self)
        new_self_weight = SelfWeight(enabled, x, y, z, LG)
        setattr(self, str(next_index), new_self_weight)
        return next_index

    def disable(self, id: int = 1) -> None:
        """Disables an instance of self-weight.

        Args:
            id (int, optional): The ID of the self-weight to disable. Defaults to 1.
        """
        self[id].disable()



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Pressure:
    def __init__(self,
                 plate_id: int = None,
                 axes: Literal["global", "local"] = 'global',
                 x_mag: float = 0,
                 y_mag: float = 0,
                 z_mag: float = 0,
                 load_group: str = None
                 ) -> None:
        """Creates an instance of the SkyCiv Pressure class.

        Args:
            plate_id (int): The plate to which the pressure is applied. Identified by the plate ID.
            axes (str, optional): The axes in which the distributed load will be applied. Defaults to 'global'. {"global" | "local"}.
            x_mag (float, optional): The magnitude of the pressure in the x-direction of the specified axes, in the units of the pressure property of the units object. Defaults to 0.
            y_mag (float, optional): The magnitude of the pressure in the y-direction of the specified axes, in the units of the pressure property of the units object. Defaults to 0.
            z_mag (float, optional): The magnitude of the pressure in the z-direction of the specified axes, in the units of the pressure property of the units object. Defaults to 0.
            load_group (str, optional): The group to which this load belongs. Defaults to None.
        """
        self.plate_id = plate_id
        self.axes = axes
        self.x_mag = x_mag
        self.y_mag = y_mag
        self.z_mag = z_mag
        self.load_group = load_group



class Settlement:
    def __init__(self,
                 node: int = None,
                 tx: float = 0,
                 ty: float = 0,
                 tz: float = 0,
                 rx: float = 0,
                 ry: float = 0,
                 rz: float = 0
                 ) -> None:
        """Creates an instance of the SkyCiv Settlement class.

        Args:
            node (int): The ID of the node at which the settlement is applied.
            tx (float, optional): Settlement displacement in the global x-axis. Defaults to None.
            ty (float, optional): Settlement displacement in the global y-axis. Defaults to None.
            tz (float, optional): Settlement displacement in the global z-axis. Defaults to None.
            rx (float, optional): Settlement rotation about the global x-axis. Defaults to None.
            ry (float, optional): Settlement rotation about the global y-axis. Defaults to None.
            rz (float, optional): Settlement rotation about the global z-axis. Defaults to None.
        """
        self.node = node
        self.tx = tx
        self.ty = ty
        self.tz = tz
        self.rx = rx
        self.ry = ry
        self.rz = rz



import copy
import types
from typing import List


def clone(dict: dict) -> dict:
    """Create a deep clone of a dictionary or array.

    Args:
        dict (dict): The dictionary to clone.

    Returns:
        dict: A clone of the input dictionary.
    """
    return copy.deepcopy(dict)


def next_object_key(dict: dict) -> int:
    """Gets next available numerical key in an object starting from 1.

    Args:
        dict (dict): The dictionary of which to get the next available key.

    Returns:
        int: The next a available integer.
    """
    nextKey = 1

    keys = dir(dict)

    while str(nextKey) in keys:
        nextKey += 1

    return nextKey


def has_get_method(dict: dict) -> bool:
    """Check if custom get method is present

    Args:
        dict (dict): The object to check.

    Returns:
        bool: Whether there is a custom get method.
    """
    get_method = getattr(dict, "get", None)
    if callable(get_method):
        if get_method != None and not isinstance(get_method, types.BuiltinFunctionType):
            return True
    return False


def keyvals(obj: dict) -> List[list]:
    """Return the key value pairs of a class or dictionary.
    Args:
        obj (dict): A class or dictionary.

    Returns:
        list[list]: An array of keyvalue pairs
    """

    if type(obj) is dict:
        return obj.items()
    else:
        return vars(obj).items()


def keys(obj: dict) -> List[list]:
    """Return the keys of a class or dictionary.
    Args:
        obj (dict): A class or dictionary.

    Returns:
        list[list]: An array of keys
    """

    if type(obj) is dict:
        return obj.keys()
    else:
        return vars(obj).keys()



from skyciv.utils.helpers import clone


class AuthObject:
    def __init__(self, username: str = None, key: str = None, session_id: str = None):
        """Creates an instance of the SkyCiv AuthObject class.

        Args:
            username (str, optional): Your SkyCiv username. Defaults to None.
            key (str, optional): Your SkyCiv API key. Defaults to None.
            session_id (str, optional): The session_id. This package will automatically add this to the APIObject if you use the APIObject's .request() method. Defaults to None.
        """
        self.username = username
        self.key = key
        self.session_id = session_id

    def get(self):
        return clone(self.__dict__)



class Node:

    def __init__(self, x: float = None, y: float = None, z: float = None) -> None:
        """Creates an instance of the SkyCiv Node class.

        Args:
            x (float): The X coordinate of the node.
            y (float): The Y coordinate of the node.
            z (float): The Z coordinate of the node.
        """
        self.x = x
        self.y = y
        self.z = z



from skyciv.classes.model.components.pressures.pressure import Pressure
from skyciv.utils.helpers import next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Pressures(ModelCollectionComponent):
    """Creates an instance of the SkyCiv Pressures class.
    """

    def add(self,
            plate_id: int,
            axes: Literal["global", "local"] = 'global',
            x_mag: float = 0,
            y_mag: float = 0,
            z_mag: float = 0,
            load_group: str = None
            ) -> int:
        """Create a pressure with the next available ID.

        Args:
            plate_id (int): The plate to which the pressure is applied. Identified by the plate ID.
            axes (str, optional): The axes in which the distributed load will be applied. Defaults to 'global'. {"global" | "local"}.
            x_mag (float, optional): The magnitude of the pressure in the x-direction of the specified axes, in the units of the pressure property of the units object. Defaults to 0.
            y_mag (float, optional): The magnitude of the pressure in the y-direction of the specified axes, in the units of the pressure property of the units object. Defaults to 0.
            z_mag (float, optional): The magnitude of the pressure in the z-direction of the specified axes, in the units of the pressure property of the units object. Defaults to 0.
            load_group (str, optional): The group to which this load belongs. Defaults to None.

        Returns:
            int: The ID of the created pressure.
        """
        next_index = next_object_key(self)

        pressure = Pressure(plate_id,
                            axes,
                            x_mag,
                            y_mag,
                            z_mag,
                            load_group
                            )
        setattr(self, str(next_index), pressure)
        return next_index

    def id_from_plate_id(self, plate_id: int) -> int:
        """Find a pressure's ID from the plate ID which it is located.

        Args:
            plate_id (int): The plate ID of the pressure to find.

        Returns:
            int: The IF of the located pressure.
        """
        found_id = None

        for k, v in vars(self).items():
            if (v["plate_id"] == plate_id):
                found_id = k

        return found_id



class MeshedPlate:
    def __init__(self,
                 parent_plate: int = None,
                 node_A: int = None,
                 node_B: int = None,
                 node_C: int = None,
                 node_D: int = None,
                 rotZ: float = 0
                 ) -> None:
        """Creates an instance of the SkyCiv MeshedPlate class.

        Args:
            parent_plate (int): The ID of the plate which this meshed plate originated from. Must refer to a plate in the plates object.
            node_A (int): The first node of the meshed plate.
            node_B (int): The second node of the meshed plate.
            node_C (int): The third node of the meshed plate.
            node_D (int, optional): The fourth node of the meshed plate. Set this to None if the meshed plate is triangular. Defaults to None.
            rotZ (float, optional): Rotation of this plate about the plate's local z-axis, in degrees. Defaults to 0.
        """
        self.parent_plate = parent_plate
        self.node_A = node_A
        self.node_B = node_B
        self.node_C = node_C
        self.node_D = node_D
        self.rotZ = rotZ



from skyciv.classes.model.components.moments.moment import Moment
from skyciv.utils.helpers import next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

class Moments(ModelCollectionComponent):
    """Creates an instance of the SkyCiv Moments class.
    """

    def add(self,
            type: Literal["n", "m"],
            node: int = None,
            member: int = None,
            position: float = None,
            x_mag: float = 0,
            y_mag: float = 0,
            z_mag: float = 0,
            load_group: str = None
            ) -> int:
        """Create a moment with the next available ID.

        Args:
            type (str): The type of object to which the load is applied. node, member. {"n" | "m"}.
            node (int, optional): The node ID which the moment is located. If type is "m", provide value None. Defaults to None.
            member (int, optional): The member ID which the moment is located. If type is "n", provide value None. Defaults to None.
            position (float, optional): The percentage from node_A to node_B of the member which the moment is located. E.g. 10 for 10 %. If type is "n", provide value None. Defaults to None.
            x_mag (float, optional): The magnitude of the moment about the x-axis. Positive = counter-clockwise, negative = clockwise. Defaults to 0.
            y_mag (float, optional): The magnitude of the moment about the y-axis. Positive = counter-clockwise, negative = clockwise. Defaults to 0.
            z_mag (float, optional): The magnitude of the moment about the z-axis. Positive = counter-clockwise, negative = clockwise. Defaults to 0.
            load_group (str, optional): The load group to which the moment will be grouped. Defaults to None.

        Returns:
            int: The ID of the created moment.
        """

        next_index = next_object_key(self)

        moment = Moment(type, node, member, position,
                        x_mag, y_mag, z_mag, load_group)
        setattr(self, str(next_index), moment)
        return next_index

    def id_from_element_id(self, type: Literal["n", "m"], element_id: int) -> int:
        """Find a moment's ID from the node ID which it is located.

        Args:
            type (str): The type of object to which the load is applied. node, member. {"n" | "m"}.
            element_id (int): The node or member ID of the moment to find.

        Returns:
            int: The ID of the found moment.
        """
        found_id = None

        elements = []

        for k, v in vars(self).items():
            if v.type == type:
                elements.append(k)

        key = "node"
        if type == "m":
            key = 'member'

        for k in elements:
            v = self[k]
            if v[key] == element_id:
                found_id = k

        return found_id

    def verify(self, loadIndex: int, type: Literal["n", "m"], node: int, member: int):
        """Check if a node or member ID is provided depending on type = "n" | "m".

        Args:
            loadIndex (int): Index of the load to check.
            type (str): Node or member. {"n" | "m"}.
            node (int): Node ID if type="n".
            member (int): Node ID if type="m".
        """
        if (type == 'm' and (not member) and node):
            print(
                f'A member ID was not provided for a moment({loadIndex}) with type="m".')

        if (type == 'n' and (not node)):
            print(
                f'A node ID was not provided for a moment({loadIndex}) with type="n".')



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Member:
    def __init__(
        self,
        node_A: int = None,
        node_B: int = None,
        section_id: int = None,
        fixity_A: str = 'FFFFFF',
        fixity_B: str = 'FFFFFF',
        type: Literal["normal", "normal_continuous",
                      "cable", "rigid"] = 'normal',
        cable_length: float = None,
        rotation_angle: float = 0,
        offset_Ax: float = 0,
        offset_Ay: float = 0,
        offset_Az: float = 0,
        offset_Bx: float = 0,
        offset_By: float = 0,
        offset_Bz: float = 0
    ) -> None:
        """Creates an instance of the SkyCiv Member class.

        Args:
            node_A (int): The ID of the start node.
            node_B (int): The ID of the end node.
            section_id (int): The ID of the section that should be applied to the member.
            fixity_A (str, optional): See docs for restraint code https://skyciv.com/api/v3/docs/s3d-model#restraint-code. Defaults to 'FFFFFF'.
            fixity_B (str, optional): See docs for restraint code https://skyciv.com/api/v3/docs/s3d-model#restraint-code. Defaults to 'FFFFFF'.
            type (str, optional): Defaults to 'normal'. {"normal" | "normal_continuous" | "cable" | "rigid"}.
            cable_length (float, optional): Required only when type = cable. Defaults to None.
            rotation_angle (float, optional): Rotation of the member about its own axis, in degrees.
            offset_Ax (float, optional): The local x distance that the member is offset from its centroid at node A.
            offset_Ay (float, optional): The local y distance that the member is offset from its centroid at node A.
            offset_Az (float, optional): The local z distance that the member is offset from its centroid at node A.
            offset_Bx (float, optional): The local x distance that the member is offset from its centroid at node B.
            offset_By (float, optional): The local y distance that the member is offset from its centroid at node B.
            offset_Bz (float, optional): The local z distance that the member is offset from its centroid at node B.
        """
        self.node_A = node_A
        self.node_B = node_B
        self.section_id = section_id
        self.fixity_A = fixity_A
        self.fixity_B = fixity_B
        self.type = type
        self.cable_length = cable_length
        self.rotation_angle = rotation_angle
        self.offset_Ax = offset_Ax
        self.offset_Ay = offset_Ay
        self.offset_Az = offset_Az
        self.offset_Bx = offset_Bx
        self.offset_By = offset_By
        self.offset_Bz = offset_Bz



from typing import List


class Section:

    def clear_self(self) -> None:
        """Delete all properties of the section.
        """
        for key in self.__dict__.keys():
            del self[key]

    def load_section_from_library(self, path: List[str], material_id: int) -> None:
        """Set the section from the SkyCiv section library.

        Args:
            path (list[str]): Provided as an array of 4 strings(see example below). It is the path of the section in the section library, obtained by inspection from within SkyCiv Section Builder or by attaining the library tree via S3D.SB.getLibraryTree.
            material_id (int): The ID of the material that is assigned to the section.

        Example::

            path = ["American", "AISC", "W shapes", "W14x22"]
        """
        self.clear_self()
        self.load_section = path
        self.material_id = material_id

    def load_custom_from_library(self, name: str, material_id: int) -> None:
        """Set the section to a custom shape created in SkyCiv section builder.

        Args:
            name (str): The name of a custom shape defined in SkyCiv Section Builder.
            material_id (int): The ID of the material that is assigned to the section.
        """
        self.clear_self()
        self.load_custom = name
        self.material_id = material_id



from skyciv.classes.model.components.supports.support import Support
from skyciv.utils.helpers import keyvals, next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent


class Supports(ModelCollectionComponent):
    """Creates an instance of the SkyCiv Supports class.
    """

    def add(self,
            node: int,
            restraint_code: str = 'FFFFFF',
            tx: float = 0,
            ty: float = 0,
            tz: float = 0,
            rx: float = 0,
            ry: float = 0,
            rz: float = 0
            ) -> int:
        """Create a support with the next available ID.

        Args:
            node (int): The ID of the node at which the support is located.
            restraint_code (str, optional): A 6 character restraint code. See docs for restraint code https://skyciv.com/api/v3/docs/s3d-model#restraint-code. Defaults to 'FFFFFF'. Defaults to 'FFFFFF'.
            tx (float, optional): Spring stiffness for translation in the x-axis. Applies only if the restraint code character for translation in x = S. Defaults to 0.
            ty (float, optional): Spring stiffness for translation in the y-axis. Applies only if the restraint code character for translation in y = S. Defaults to 0.
            tz (float, optional): Spring stiffness for translation in the z-axis. Applies only if the restraint code character for translation in z = S. Defaults to 0.
            rx (float, optional): Spring stiffness for rotation about the x-axis. Applies only if the restraint code character for rotation about x = S. Defaults to 0.
            ry (float, optional): Spring stiffness for rotation about the y-axis. Applies only if the restraint code character for rotation about y = S. Defaults to 0.
            rz (float, optional): Spring stiffness for rotation about the z-axis. Applies only if the restraint code character for rotation about z = S. Defaults to 0.

        Returns:
            int: The ID of the created support.
        """

        next_index = next_object_key(self)

        support = Support(node, restraint_code, tx, ty, tz, rx, ry, rz)
        setattr(self, str(next_index), support)
        return next_index

    def id_from_node_id(self, node_id: int) -> int:
        """Find a support's ID from the node ID which it is located.

        Args:
            node_id (int): The node ID of the support to find.

        Returns:
            int: The ID of the located support.
        """
        id = None
        for k, v in keyvals(self):
            if (v.node == node_id):
                id = k

        return id




from skyciv.constants.skyciv_sections import SkyCivSections
from skyciv.classes.api_object.api_object import ApiObject
from skyciv.classes.model.model import Model
from skyciv.lib.request import request

sections = SkyCivSections



from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent


class LoadCombination(ModelCollectionComponent):

    def __init__(self, name: str = None, combination_factors: dict = None) -> None:
        """Creates an instance of the SkyCiv LoadCombination class.

        Args:
            name (str): The name of the load combination.
            combination_factors (dict): Key value pairs for the factors to apply to the load groups.
        """
        self.set(name, combination_factors)

    def set(self, name: str, combination_factors: dict) -> None:
        """Set individual properties of the LoadCombination object.

        Args:
            name (str): The name of the load combination.
            combination_factors (dict): An object of key value pairs specifying the load group factor.

        Raises:
            Exception: If a combination factor key is "name" as this is reserved.

        Example::

            lc = LoadCombination()

            factors = {
                "SW": 1,
                "windCase": 1,
                "liveLoad": 1.5
            }

            lc.clear_all() # To remove any existing factors.
            lc.set("LC1", factors)
        """
        self.name = name

        if not combination_factors == None:
            for k, v in combination_factors.items():
                if k == "name":
                    raise Exception(
                        "A load combination can not have the name 'name' as this is a reserved keyword.")
                else:
                    setattr(self, str(k), v)





class SkyCivFunctions():
    S3D_session_start = "S3D.session.start"
    S3D_model_set = "S3D.model.set"
    S3D_model_get = "S3D.model.get"
    S3D_model_repair = "S3D.model.repair"
    S3D_model_solve = "S3D.model.solve"
    S3D_model_takeScreenshot = "S3D.model.takeScreenshot"
    S3D_model_mesh = "S3D.model.mesh"
    S3D_results_set = "S3D.results.set"
    S3D_results_get = "S3D.results.get"
    S3D_results_fetchMemberResult = "S3D.results.fetchMemberResult"
    S3D_results_getAnalysisReport = "S3D.results.getAnalysisReport"
    S3D_design_member_getInput = "S3D.design.member.getInput"
    S3D_design_member_check = "S3D.design.member.check"
    S3D_design_member_optimize = "S3D.design.member.optimize"
    standalone_member_check = "standalone.member.check"
    S3D_design_rc_getInput = "S3D.design.rc.getInput"
    S3D_design_rc_check = "S3D.design.rc.check"
    S3D_SB_loadLibraryShape = "S3D.SB.loadLibraryShape"
    S3D_SB_getLibraryTree = "S3D.SB.getLibraryTree"
    S3D_SB_buildCustomShape = "S3D.SB.buildCustomShape"
    S3D_SB_solve = "S3D.SB.solve"
    S3D_SB_runGSD = "S3D.SB.runGSD"
    S3D_file_save = "S3D.file.save"
    S3D_file_open = "S3D.file.open"
    S3D_file_share = "S3D.file.share"
    S3D_file_getFileDirectory = "S3D.file.getFileDirectory"



from skyciv.classes.functions.api_function import ApiFunction
from skyciv.classes.model.components.load_combinations.load_combinations import LoadCombinations
from skyciv.utils.helpers import clone, has_get_method, keyvals
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

class ApiFunctions:

    def __init__(self):
        """Creates an instance of the SkyCiv ApiFunctions class.
        """
        self.function_list = []

    def get(self):
        """Get the functions array.
        """
        list = clone(self.function_list)

        # Run get method on functions
        for i in range(len(list)):
            list[i] = list[i].get()

            # Run get method on any function args if they have the method available
            for k, v in list[i]["arguments"].items():
                if has_get_method(v):
                    list[i]["arguments"][k] = v.get()

        return list

    def add(self, function_name: Literal["S3D.session.start", "S3D.model.set", "S3D.model.get", "S3D.model.repair", "S3D.model.solve", "S3D.model.takeScreenshot", "S3D.model.mesh", "S3D.results.set", "S3D.results.get", "S3D.results.fetchMemberResult", "S3D.results.getAnalysisReport", "S3D.model.takeScreenshot", "S3D.design.member.getInput", "S3D.design.member.check", "S3D.design.member.optimize", "standalone.member.check", "S3D.design.rc.getInput", "S3D.design.rc.check", "S3D.SB.loadLibraryShape", "S3D.SB.getLibraryTree", "S3D.SB.buildCustomShape", "S3D.SB.solve", "S3D.SB.runGSD", "S3D.file.save", "S3D.file.open", "S3D.file.share", "S3D.file.getFileDirectory"], args: dict = {}) -> None:
        """Add a SkyCiv function to the ApiObject's functions list. See the docs for available functions and respective arguments. https://skyciv.com/api/v3/docs/the-request-object#functions.
        Args:
            function_name (str): The SkyCiv function name. {"S3D.session.start" | "S3D.model.set" | "S3D.model.get" | "S3D.model.repair" | "S3D.model.solve" | "S3D.model.takeScreenshot" | "S3D.model.mesh" | "S3D.results.set" | "S3D.results.get" | "S3D.results.fetchMemberResult" | "S3D.results.getAnalysisReport" | "S3D.model.takeScreenshot" | "S3D.design.member.getInput" | "S3D.design.member.check" | "S3D.design.member.optimize" | "standalone.member.check" | "S3D.design.rc.getInput" | "S3D.design.rc.check" | "S3D.SB.loadLibraryShape" | "S3D.SB.getLibraryTree" | "S3D.SB.buildCustomShape" | "S3D.SB.solve" | "S3D.SB.runGSD" | "S3D.file.save" | "S3D.file.open" | "S3D.file.share" | "S3D.file.getFileDirectory"}
            args (dict, optional): See the docs for the available arguments. Defaults to {}.
        """
        fn = ApiFunction(function_name, args)

        # Where there is an s3d_model, overwrite load_combinations with its get method it doesnt have the list prop.
        for arg_key, arg_value in args.items():
            if (arg_key == 's3d_model'):
                for k, v in keyvals(arg_value):
                    if (isinstance(v, LoadCombinations)):
                        # args["s3d_model"]["load_combinations"] = v.get()
                        args["s3d_model"].set({
                            "load_combinations": v.get(),
                        })

        self.function_list.append(fn)

    def remove(self, function_name: Literal["S3D.session.start", "S3D.model.set", "S3D.model.get", "S3D.model.repair", "S3D.model.solve", "S3D.model.takeScreenshot", "S3D.model.mesh", "S3D.results.set", "S3D.results.get", "S3D.results.fetchMemberResult", "S3D.results.getAnalysisReport", "S3D.model.takeScreenshot", "S3D.design.member.getInput", "S3D.design.member.check", "S3D.design.member.optimize", "standalone.member.check", "S3D.design.rc.getInput", "S3D.design.rc.check", "S3D.SB.loadLibraryShape", "S3D.SB.getLibraryTree", "S3D.SB.buildCustomShape", "S3D.SB.solve", "S3D.SB.runGSD", "S3D.file.save", "S3D.file.open", "S3D.file.share", "S3D.file.getFileDirectory"]) -> None:
        """Remove all occurences of a SkyCiv function from the ApiObject's functions list. See the docs for available functions and respective arguments. https://skyciv.com/api/v3/docs/the-request-object#functions

        Args:
            function_name (str): The SkyCiv function name. {"S3D.session.start" | "S3D.model.set" | "S3D.model.get" | "S3D.model.repair" | "S3D.model.solve" | "S3D.model.takeScreenshot" | "S3D.model.mesh" | "S3D.results.set" | "S3D.results.get" | "S3D.results.fetchMemberResult" | "S3D.results.getAnalysisReport" | "S3D.model.takeScreenshot" | "S3D.design.member.getInput" | "S3D.design.member.check" | "S3D.design.member.optimize" | "standalone.member.check" | "S3D.design.rc.getInput" | "S3D.design.rc.check" | "S3D.SB.loadLibraryShape" | "S3D.SB.getLibraryTree" | "S3D.SB.buildCustomShape" | "S3D.SB.solve" | "S3D.SB.runGSD" | "S3D.file.save" | "S3D.file.open" | "S3D.file.share" | "S3D.file.getFileDirectory"}.
        """
        new_list = []
        for fn in self.function_list:
            if fn["function"] != function_name:
                new_list.append(fn)

        self.function_list = new_list

    def remove_all(self) -> None:
        """Remove all SkyCiv functions from the ApiObject's functions list. See the docs for available functions and respective arguments. https://skyciv.com/api/v3/docs/the-request-object#functions
        """
        self.function_list = []



from skyciv.classes.model.components.sections.section import Section
from skyciv.utils.helpers import next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent
from typing import List


class Sections(ModelCollectionComponent):
    """Creates an instance of the SkyCiv Sections class.
    """

    def add_library_section(self, path: List[str], material_id: int) -> int:
        """Add a section from the SkyCiv section library.

        Args:
            path (list[str]): Provided as an array of 4 strings (see example below). It is the path of the section in the section library, obtained by inspection from within SkyCiv Section Builder or by attaining the library tree via S3D.SB.getLibraryTree.
            material_id (int): The ID of the material that is assigned to the section.

        Returns:
            int: The ID of the new section.

        Example::

            path = ["American", "AISC", "W shapes", "W14x22"]
        """
        next_index = next_object_key(self)
        new_section = Section()
        new_section.load_section_from_library(path, material_id)
        setattr(self, str(next_index), new_section)
        return next_index

    def add_custom_section(self, name: str, material_id: int) -> int:
        """Add a custom section created in SkyCiv section builder.

        Args:
            name (str): The name of a custom shape defined in SkyCiv Section Builder.
            material_id (int): The ID of the material that is assigned to the section.

        Returns:
            int: The ID of the new section.
        """
        next_index = next_object_key(self)
        new_section = Section()
        new_section.load_custom_from_library(name, material_id)
        setattr(self, str(next_index), new_section)
        return next_index

    def ids_from_path(self, path: List[str]) -> List[int]:
        """Get all sections that match a SkyCiv section library path.

        Args:
            path (list[str]): Provided as an array of 4 strings(see example below). It is the path of the section in the section library, obtained by inspection from within SkyCiv Section Builder or by attaining the library tree via S3D.SB.getLibraryTree.

        Returns:
            [type]: A list of IDs that match the input path.

        Example::

            path = ["American", "AISC", "W shapes", "W14x22"]
        """
        ids = []

        for k, v in vars(self).items():
            if (v.path == path):
                ids.append(k)

        if (len(ids) == 0):
            ids = None

        return ids



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

class Moment:

    def __init__(
        self,
        type: Literal["n", "m"] = None,
        node: int = None,
        member: int = None,
        position: float = None,
        x_mag: float = 0,
        y_mag: float = 0,
        z_mag: float = 0,
        load_group: str = None
    ) -> None:
        """Creates an instance of the SkyCiv Moment class

        Args:
            type (str): The type of object to which the load is applied. node, member. {"n" | "m"}.
            node (int, optional): The node ID which the moment is located. If type is "m", provide value None. Defaults to None.
            member (int, optional): The member ID which the moment is located. If type is "n", provide value None. Defaults to None.
            position (float, optional): The percentage from node_A to node_B of the member which the moment is located. E.g. 10 for 10 %. If type is "n", provide value None. Defaults to None.
            x_mag (float, optional): The magnitude of the moment about the x-axis. Positive = counter-clockwise, negative = clockwise. Defaults to 0.
            y_mag (float, optional): The magnitude of the moment about the y-axis. Positive = counter-clockwise, negative = clockwise. Defaults to 0.
            z_mag (float, optional): The magnitude of the moment about the z-axis. Positive = counter-clockwise, negative = clockwise. Defaults to 0.
            load_group (str, optional): The load group to which the moment will be grouped. Defaults to None.

        Example::

            node_load = Moment('n', 1, None, None, 0, -5.3, 0, "LG1")
            member_load = Moment('m', None, 3, 32.4, 0, -5.3, 0, "LG1")
        """
        self.type = type
        self.node = node
        self.member = member
        self.position = position
        self.x_mag = x_mag
        self.y_mag = y_mag
        self.z_mag = z_mag
        self.load_group = load_group



import setuptools

with open("./README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="skyciv",
    version="2.0.4",
    description="A simplified way to use the SkyCiv API with Python.",
    py_modules=["skyciv"],
    package_dir={"": "src"},
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="skyciv structural analysis design API AISC Eurocode CSA steel concrete BIM",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    extras_require={
        "dev": [
            "pytest>=3.7"
        ]
    },
    url="https://github.com/skyciv/skyciv-pip",
    author="Steve Richardson",
    author_email="steve.richardson@skyciv.com",
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=['typing-extensions']
)



from skyciv.classes.model.components.point_loads.point_load import PointLoad
from skyciv.utils.helpers import next_object_key
from skyciv.classes.model.components._base_class.model_collection_component import ModelCollectionComponent
from typing import List
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

class PointLoads(ModelCollectionComponent):
    """Creates an instance of the SkyCiv PointLoads class.
    """

    def add(self,
            type: Literal["n", "m"],
            node: int = None,
            member: int = None,
            position: float = None,
            x_mag: float = 0,
            y_mag: float = 0,
            z_mag: float = 0,
            load_group: str = 'LG1'
            ) -> int:
        """Create a point load with the next available ID.

        Args:
            type (str): The type of object to which the load is applied. node, member. {"n" | "m"}.
            node (int, optional): The node ID which the point load is located. If type is "m", provide value None. Defaults to None.
            member (int, optional): The member ID which the point load is located. If type is "n", provide value None. Defaults to None.
            position (float, optional): The percentage from node_A to node_B of the member which the point load is located. E.g. 10 for 10 %. If type is "n", provide value None. Defaults to None.
            x_mag (float, optional): The magnitude of the point load force along the x-axis. Defaults to 0.
            y_mag (float, optional): The magnitude of the point load force along the y-axis. Defaults to 0.
            z_mag (float, optional): The magnitude of the point load force along the z-axis. Defaults to 0.
            load_group (str, optional): The load group to which the point load will be grouped. Defaults to 'LG1'.

        Returns:
            int: The ID of the created point load.
        """
        next_index = next_object_key(self)

        self.verify(next_index, type, node, member)

        pl = PointLoad(
            type,
            node,
            member,
            position,
            x_mag,
            y_mag,
            z_mag,
            load_group
        )
        setattr(self, str(next_index), pl)
        return next_index

    def id_from_element_id(self, type: Literal["n", "m"], element_id: int) -> int:
        """Find a point load's ID from the node ID which it is located.

        Args:
            type (str): The type of object to which the load is applied. node, member. {"n" | "m"}.
            element_id (int): The node or member ID of the point load to find.

        Returns:
            int: The ID of the located point load.
        """

        found_id = None
        elements = []

        for k, v in vars(self).items():
            if v.type == type:
                elements.append(k)

        key = "node"
        if type == "m":
            key = "member"

        for k in elements:
            v = self[k]
            if (v[key] == element_id):
                found_id = k

        return found_id

    def verify(self, loadIndex: int, type: Literal["n", "m"], node: int, member: int):
        """Check if a node or member ID is provided depending on type = "n" | "m".

        Args:
            loadIndex (int): Index of the load to check.
            type (str): Node or member. {"n" | "m"}.
            node (int): Node ID if type="n".
            member (int): Node ID if type="m".
        """
        if (type == 'm' and (not member) and node):
            print(
                f'A member ID was not provided for a point load({loadIndex}) with type="m".')

        if (type == 'n' and (not node)):
            print(
                f'A node ID was not provided for a point load({loadIndex}) with type="n".')



from typing import List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class AreaLoad:
    def __init__(self,
                 type: Literal["one_way", "two_way",
                               "column_wind_load", "open_structure"] = None,
                 nodes: List[int] = None,
                 mag: float = None,
                 direction: Literal["X", "Y", "Z"] = None,
                 elevations: str = None,
                 mags: str = None,
                 column_direction: str = None,
                 loaded_members_axis: Literal["all", "major"] = None,
                 LG: str = None
                 ) -> None:
        """Creates an instance of the SkyCiv AreaLoad class.

        Args:
            type (str): How the area load should distribute the load. {"one_way" | "two_way" | "column_wind_load" | "open_structure"}.
            nodes (list[int]): The IDs of the nodes which define the area for loading. Specify 3 or 4 values, in sequential order(clockwise or counterclockwise direction).
            mag (float): The magnitude of the load, in the units of pressure.
            direction (str): The direction of the load in the global axes. {"X" | "Y" | "Z"}.
            elevations (str, optional): Relevant only if "type": "column_wind_load". The elevations between which the corresponding pressure magnitudes (see next row in this table) should be applied. This property should have 1 more value than the corresponding pressure magnitudes property. Defaults to None.
            mags (str, optional): Relevant only if "type": "column_wind_load". The magnitudes of pressures which should be applied between the corresponding elevations (see above row in this table). This property should have 1 less value than the corresponding elevations property. Defaults to None.
            column_direction (str, optional): Relevant only if "type": "one_way" or "type": "column_wind_load". The span direction of the applied area load. The values must be the IDs of 2 nodes which are in the nodes property. Defaults to None.
            loaded_members_axis (str, optional): Relevant only if "type": "open_structure".Whether to apply the open structure load to all members attaching to the nodes (indicated by all), or to only those members which lie along the global XYZ axes(indicated by major). Defaults to None. {"all" | "major"}
            LG (str, optional): The load group to which this area load belongs. Defaults to None.
        """

        self.type = type
        self.nodes = nodes
        self.mag = mag
        self.direction = direction
        self.elevations = elevations
        self.mags = mags
        self.column_direction = column_direction
        self.loaded_members_axis = loaded_members_axis
        self.LG = LG




import json
from skyciv.classes.auth.auth_object import AuthObject
from skyciv.classes.functions.api_functions import ApiFunctions
from skyciv.classes.options.options_object import OptionsObject
from skyciv.utils.helpers import clone, keyvals
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from skyciv.lib.request import request


class ApiObject:
    def __init__(self):
        """Create an instance of the SkyCiv API Object.
        """
        self.auth = AuthObject()
        self.options = OptionsObject()
        self.functions = ApiFunctions()

    def request(self, http_or_https: Literal["https", "http"] = 'https', version: int = 3):
        """Send the ApiObject to the SkyCiv API.

        Args:
            http_or_https (str, optional): Whether to use http or https. {"http" | "https"}. Defaults to 'https'.
            version (int, optional): API version. {2 | 3}. Defaults to 3.
        """
        res = request(self.to_json(), {http_or_https, version})
        if "response" in res:
            if "last_session_id" in res["response"]:
                self.auth.session_id = res["response"]["last_session_id"]

        return res

    def get(self) -> dict:
        """Converts the ApiObject into a Python dictionary that will parse to JSON in the SkyCiv format.

        Returns:
            dict: The API object as a Python dictionary.
        """
        api_object = {
            "auth": clone(self.auth.get()),
            "options": clone(self.options.get()),
            "functions": clone(self.functions.get()),
        }
        return api_object

    def to_json(self) -> str:
        model = self.get()
        return json.dumps(model)
